import os
import shutil
from tarfile import TarFile
from typing import Callable, List, Optional, IO
from zipfile import ZipFile

import sfu_torch_lib.io as io


IsMemberType = Callable[[str], bool]


class FileFetcher:
    members: List[str]

    def __init__(self, path: str, is_member: IsMemberType) -> None:
        self.path = path
        self.is_member = is_member

    def __len__(self) -> int:
        return len(self.members)

    def __getitem__(self, index: int) -> str:
        return self.members[index]

    def open_member(self, key: str) -> IO:
        raise NotImplementedError()


class ZipFileFetcher(FileFetcher):
    def __init__(
            self, path: str,
            is_member: IsMemberType,
            ephemeral_path: Optional[str] = os.environ.get('DATA_EPHEMERAL_PATH'),
    ) -> None:

        super().__init__(path, is_member)

        self.ephemeral_path = ephemeral_path

        self.members = self.get_members(path, is_member)
        self.container: Optional[ZipFile] = None

    @staticmethod
    def get_members(path: str, is_member: Callable[[str], bool]) -> List[str]:
        return [name for name in ZipFile(path).namelist() if is_member(name)]

    def open_member(self, key: str) -> IO:
        if self.container is None:
            self.container = ZipFile(self.path)

        if self.ephemeral_path is not None:
            prefix = io.generate_path(prefix=self.ephemeral_path)
            path = os.path.join(prefix, key)

            if not io.exists(path):
                os.makedirs(os.path.dirname(path), exist_ok=True)
                self.container.extract(key, prefix)

            file_object = io.open(path)

        else:
            file_object = self.container.open(key)

        return file_object


class TarFileFetcher(FileFetcher):
    def __init__(
            self, path: str,
            is_member: IsMemberType,
            ephemeral_path: Optional[str] = os.environ.get('DATA_EPHEMERAL_PATH'),
    ) -> None:

        super().__init__(path, is_member)

        self.ephemeral_path = ephemeral_path

        self.members = self.get_members(path, is_member)
        self.container: Optional[TarFile] = None

    @staticmethod
    def get_members(path: str, is_member: Callable[[str], bool]) -> List[str]:
        return [name for name in TarFile(path).getnames() if is_member(name)]

    def open_member(self, key: str) -> IO:
        if self.container is None:
            self.container = TarFile(self.path)

        if self.ephemeral_path is not None:
            prefix = io.generate_path(prefix=self.ephemeral_path)
            path = os.path.join(prefix, key)

            if not io.exists(path):
                os.makedirs(os.path.dirname(path), exist_ok=True)
                self.container.extract(key, prefix)

            file_object = io.open(path)

        else:
            file_object_or_none = self.container.extractfile(key)

            if file_object_or_none is None:
                raise IOError(f'Member {key} is not a file.')

            file_object = file_object_or_none

        return file_object


class FileSystemFileFetcher(FileFetcher):
    def __init__(
            self,
            path: str,
            is_member: IsMemberType,
            ephemeral_path: Optional[str] = os.environ.get('DATA_EPHEMERAL_PATH'),
    ) -> None:

        super().__init__(path, is_member)

        self.ephemeral_path = ephemeral_path

        self.members = self.get_members(path, is_member)

    @staticmethod
    def get_members(path: str, is_member: Callable[[str], bool]) -> List[str]:
        return [
            name
            for name
            in (os.path.relpath(key, path) for key in io.get_files(path))
            if is_member(name)
        ]

    def open_member(self, key: str) -> IO:
        path_original = io.generate_path(key, self.path)

        if self.ephemeral_path is not None:
            prefix = io.generate_path(prefix=self.ephemeral_path)
            path_ephemeral = os.path.join(prefix, key)

            if not io.exists(path_ephemeral):
                os.makedirs(os.path.dirname(path_ephemeral), exist_ok=True)
                shutil.copy(path_original, path_ephemeral)

            file_object = io.open(path_ephemeral)

        else:
            file_object = io.open(path_original)

        return file_object


def get_file_fetcher(path: str, is_member: IsMemberType = lambda _: True) -> FileFetcher:
    if path.endswith('.zip'):
        return ZipFileFetcher(path, is_member)
    elif path.endswith('.tar'):
        return TarFileFetcher(path, is_member)
    else:
        return FileSystemFileFetcher(path, is_member)



from torch.nn import Conv2d, ConvTranspose2d


def conv(in_channels: int, out_channels: int, kernel_size: int, stride: int):
    return Conv2d(
        in_channels,
        out_channels,
        kernel_size=(kernel_size, kernel_size),
        stride=(stride, stride),
        padding=(kernel_size // 2, kernel_size // 2),
    )


def deconv(in_channels: int, out_channels: int, kernel_size: int, stride: int):
    return ConvTranspose2d(
        in_channels,
        out_channels,
        kernel_size=(kernel_size, kernel_size),
        stride=(stride, stride),
        output_padding=(stride - 1, stride - 1),
        padding=(kernel_size // 2, kernel_size // 2),
    )



import os
from typing import Type

import mlflow
import torch
from pytorch_lightning import LightningModule

import sfu_torch_lib.io as io


def load_model(run_id: str, module_class: Type[LightningModule], filename: str = 'last') -> LightningModule:
    run = mlflow.get_run(run_id)

    model_path = os.path.join(run.info.artifact_uri, f'{filename}.ckpt')

    with io.open(model_path) as model_file:
        model = module_class.load_from_checkpoint(model_file)

    return model


def load_state(run_id: str, model: LightningModule, filename: str = 'last') -> None:
    device = torch.device('cuda') if torch.cuda.device_count() else torch.device('cpu')

    run = mlflow.get_run(run_id)

    model_path = os.path.join(run.info.artifact_uri, f'{filename}.ckpt')

    with io.open(model_path) as model_file:
        checkpoint = torch.load(model_file, device)

    model.load_state_dict(checkpoint['state_dict'])



import datetime
import functools
import json
import os
import socket
import traceback
from typing import Sequence, Optional, Callable

import requests

import sfu_torch_lib.mlflow as mlflow
import sfu_torch_lib.parameters as parameters_lib
import sfu_torch_lib.utils as utils


DATE_FORMAT = '%Y-%m-%d %H:%M:%S'


def notify(
        function: Callable,
        webhook_url: Optional[str] = os.getenv('SLACK_URL'),
        user_ids: Optional[Sequence[str]] = None,
        override: bool = utils.to_bool_or_false(os.getenv('DEBUGGING')),
) -> Callable:
    """
    Executes a function and sends a Slack notification with the final status (successfully finished or
    crashed). Also sends a Slack notification before executing the function.
    Visit https://api.slack.com/incoming-webhooks#create_a_webhook for more details.
    Visit https://api.slack.com/methods/users.identity for more details.

    :param function: Function to annotate and execute.
    :param webhook_url: The webhook URL to your Slack channel. If missing, no notifications will be made.
    :param user_ids: Optional user ids to notify.
    :param override: Skip sending notifications.
    """
    if user_ids is None:
        user_ids = [os.environ['SLACK_USER']] if os.getenv('DEVELOPMENT_MODE') else []

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        if not webhook_url or override:
            return function(*args, **kwargs)

        user_mentions = ', '.join(f'<@{user_id}>' for user_id in user_ids)

        parameters = '\n'.join(
            f'\t{key}: {value}'
            for key, value
            in parameters_lib.get_script_parameters(function, ignore_keyword_arguments=False).items()
        )

        start_time = datetime.datetime.now()
        host_name = socket.gethostname()

        if 'RANK' in os.environ:
            master_process = (int(os.environ['RANK']) == 0)
            host_name += f' - RANK: {os.environ["RANK"]}'
        else:
            master_process = True

        if master_process:
            contents = [
                '*🎬 Your job has started*',
                f'Machine name: {host_name}',
                f'Starting time: {start_time.strftime(DATE_FORMAT)}',
                f'Parameters:\n{parameters}',
                f'User: {user_mentions}',
            ]

            message = {'text': '\n'.join(contents)}
            requests.post(webhook_url, json.dumps(message))

        try:
            value = function(*args, **kwargs)

            metrics = '\n'.join(
                f'\t{key}: {value}'
                for key, value
                in mlflow.get_metrics().items()
            )

            if master_process:
                end_time = datetime.datetime.now()
                elapsed_time = end_time - start_time
                contents = [
                    '*🎉 Your job is complete*',
                    f'Machine name: {host_name}',
                    f'Starting time: {start_time.strftime(DATE_FORMAT)}',
                    f'End date: {end_time.strftime(DATE_FORMAT)}',
                    f'Duration: {elapsed_time}',
                    f'Main call returned value: {value}',
                    f'Parameters:\n{parameters}',
                    f'Metrics:\n{metrics}',
                    f'User: {user_mentions}',
                ]

                message = {'text': '\n'.join(contents)}
                requests.post(webhook_url, json.dumps(message))

            return value

        except Exception as exception:
            end_time = datetime.datetime.now()
            elapsed_time = end_time - start_time
            contents = [
                '*☠️ Your job has crashed*',
                f'Machine name: {host_name}',
                f'Starting time: {start_time.strftime(DATE_FORMAT)}',
                f'End date: {end_time.strftime(DATE_FORMAT)}',
                f'Duration: {elapsed_time}\n\n',
                f'Error:\n{exception}\n\n',
                f'Traceback:\n{traceback.format_exc()}',
                f'User: {user_mentions}',
            ]

            message = {'text': '\n'.join(contents)}
            requests.post(webhook_url, json.dumps(message))

            raise exception

    return wrapper



import functools
from typing import Tuple, Sequence, Iterator, Optional, NamedTuple

import torch
import torch.autograd as autograd
from torch import Tensor
from torch.optim import Optimizer


class LossArguments(NamedTuple):
    target_loss: Tensor
    adaptive_loss: Optional[Tensor] = None
    scale_factor: Optional[float] = 1.


def get_parameters(optimizer: Optimizer) -> Iterator[Tensor]:
    yield from (
        parameter
        for parameter_group in optimizer.param_groups
        for _, parameters in parameter_group.items()
        for parameter in (parameters if isinstance(parameters, list) else [parameters])
        if torch.is_tensor(parameter) and parameter.requires_grad
    )


def manual_step(optimizer: Optimizer, loss: Tensor, allow_unused: bool = False) -> None:
    optimizer.zero_grad()

    parameters = list(get_parameters(optimizer))
    gradients = autograd.grad(loss, parameters, retain_graph=True, allow_unused=allow_unused)

    for parameter, gradient in zip(parameters, gradients):
        parameter.grad = gradient

    optimizer.step()


def compute_norms(adaptive_gradients: Sequence[Tensor], target_gradients: Sequence[Tensor]) -> Tuple[Tensor, Tensor]:
    adaptive_norm, target_norm = functools.reduce(
        lambda tuple_first, tuple_second: (tuple_first[0] + tuple_second[0], tuple_first[1] + tuple_second[1]),
        (
            (torch.sum(adaptive_gradient ** 2), torch.sum(target_gradient ** 2))
            for adaptive_gradient, target_gradient in zip(adaptive_gradients, target_gradients)
            if target_gradient is not None and adaptive_gradient is not None
        ),
    )

    adaptive_norm, target_norm = torch.sqrt(adaptive_norm), torch.sqrt(target_norm)

    return adaptive_norm, target_norm


def combine_gradients(
        adaptive_gradients: Sequence[Tensor],
        target_gradients: Sequence[Tensor],
        scaling_factor: Tensor,
) -> Iterator[Optional[Tensor]]:

    for adaptive_gradient, target_gradient in zip(adaptive_gradients, target_gradients):
        if adaptive_gradient is not None and target_gradient is not None:
            yield target_gradient + scaling_factor * adaptive_gradient
        elif adaptive_gradient is not None:
            yield adaptive_gradient
        elif target_gradient is not None:
            yield target_gradient
        else:
            yield None


def calculate_scaled_gradients(
        parameters: Sequence[Tensor],
        adaptive_loss: Tensor,
        target_loss: Tensor,
        scale_factor: float,
) -> Iterator[Optional[Tensor]]:

    adaptive_gradients = autograd.grad(adaptive_loss, parameters, retain_graph=True, allow_unused=True)
    target_gradients = autograd.grad(target_loss, parameters, retain_graph=True, allow_unused=True)

    adaptive_norm, target_norm = compute_norms(adaptive_gradients, target_gradients)
    scaling_factor = torch.minimum(adaptive_norm, target_norm) / adaptive_norm * scale_factor

    yield from combine_gradients(adaptive_gradients, target_gradients, scaling_factor)


def collect_gradients(parameters: Sequence[Tensor], losses: Sequence[LossArguments],) -> Iterator[Tuple[Tensor, ...]]:
    collected_gradients = []

    for loss in losses:
        if loss.adaptive_loss is None:
            gradients: Iterator[Optional[Tensor]] = (
                loss.scale_factor * gradient if gradient is not None else None
                for gradient
                in autograd.grad(loss.target_loss, parameters, retain_graph=True, allow_unused=True)
            )
        else:
            assert loss.scale_factor is not None
            gradients = calculate_scaled_gradients(parameters, loss.adaptive_loss, loss.target_loss, loss.scale_factor)

        collected_gradients.append(gradients)

    yield from (
        tuple(gradient for gradient in gradients if gradient is not None)
        for gradients in zip(*collected_gradients)
    )


def adaptive_clipping(optimizer: Optimizer, losses: Sequence[LossArguments]) -> None:
    optimizer.zero_grad()

    parameters = list(get_parameters(optimizer))
    gradients = collect_gradients(parameters, losses)

    for parameter, gradient in zip(parameters, gradients):
        if len(gradient) > 0:
            parameter.grad = sum(gradient)

    optimizer.step()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sfu_torch_lib', 'sfu_torch_lib.modules']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1,<2',
 'mlflow-skinny>=1,<2',
 'numpy>=1,<2',
 'pytorch-lightning>=1,<2',
 'requests>=2,<3',
 's3fs>=2022,<2023',
 'torch>=1,<2']

setup_kwargs = {
    'name': 'sfu-torch-lib',
    'version': '0.0.10',
    'description': 'Libraries that support the development of machine learning models in PyTorch.',
    'long_description': None,
    'author': 'Anderson de Andrade',
    'author_email': 'anderson_de_andrade@sfu.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<=3.10',
}


setup(**setup_kwargs)



import functools
import math
import re
from typing import List, Tuple, TypeVar, Generator, Type, Union, Sequence, Mapping, Callable, Generic, Dict, Optional

from torch.nn import ELU, Identity, ReLU, LeakyReLU, Sigmoid, Softmax, Tanh, Module
from torch.utils.data import Dataset


T = TypeVar('T')


class AttributeDict(Generic[T]):
    def __init__(self, dictionary: Dict[str, T]) -> None:
        self.dictionary = dictionary

        for key, value in self.items():
            setattr(self, key.lower().replace(' ', '_'), value)

    def __getattr__(self, name):
        return getattr(self, name)

    def __getitem__(self, key):
        return self.dictionary[key]

    def __repr__(self):
        return repr(self.dictionary)

    def __len__(self):
        return len(self.dictionary)

    def __contains__(self, item):
        return item in self.dictionary

    def __iter__(self):
        return iter(self.dictionary)

    def keys(self):
        return self.dictionary.keys()

    def values(self):
        return self.dictionary.values()

    def items(self):
        return self.dictionary.items()


def get_pairs(items: List[T]) -> Generator[Tuple[T, T], None, None]:
    for index in range(len(items) - 1):
        yield items[index], items[index + 1]


def create_selector(indices: Union[int, Sequence[int]]) -> Callable[[Sequence[T]], Union[T, Tuple[T, ...]]]:
    def select(items: Sequence[T]) -> Union[T, Tuple[T, ...]]:
        if isinstance(indices, int):
            return items[indices]

        return tuple(items[index] for index in indices)

    return select


def to_list(*args: Union[Sequence[T], T]) -> List[T]:
    items: List[T] = []

    for argument in args:
        if isinstance(argument, Sequence):
            items += argument
        else:
            items.append(argument)

    return items


def to_bool(value: Optional[str]) -> Optional[bool]:
    if value is None:
        return None

    if value == '0' or value.lower() == 'false':
        return False

    return bool(value)


def to_bool_or_false(value: Optional[str]) -> bool:
    return to_bool(value) or False


def get_activation(name: str) -> Type[Module]:
    return {
        'elu': ELU,
        'identity': Identity,
        'relu': ReLU,
        'leaky_relu': LeakyReLU,
        'sigmoid': Sigmoid,
        'softmax': Softmax,
        'tanh': Tanh,
    }[name]


def parse_map_sequence(string: str, sequence_type: Type[T]) -> Mapping[str, Sequence[T]]:
    substrings = re.split(r'(?<=]),', string)

    output = {}

    for substring in substrings:
        key, value = substring.split(':')

        output[key] = list(map(sequence_type, value.strip('[]').split(',')))

    return output


def get_num_steps(dataset: Dataset, batch_size: int, frequency: int = 10) -> int:
    return max(1, math.floor(len(dataset) / batch_size / frequency))  # type: ignore


parse_map_ints = functools.partial(parse_map_sequence, sequence_type=int)
parse_map_floats = functools.partial(parse_map_sequence, sequence_type=float)
parse_map_strings = functools.partial(parse_map_sequence, sequence_type=str)
